#  Copyright (c) 1997-2024
#  Ewgenij Gawrilow, Michael Joswig, and the polymake team
#  Technische Universität Berlin, Germany
#  https://polymake.org
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2, or (at your option) any
#  later version: http://www.gnu.org/licenses/gpl.txt.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#-------------------------------------------------------------------------------

package Visual::PhylogeneticTree;

use Polymake::Struct (
   [ '@ISA' => 'Visual::Graph' ],
   #[ '$NodeLabels' => '$this->create_node_labels(#%)', default =>'undef' ],
   #[ '$NodeColor' => 'unify_decor(#%)', default => 'undef' ],
   #[ '$NodeBorderColor' => 'unify_decor(#%)', default => '"0 0 0"' ],
);

sub create_node_labels {
   
}

1;

# Local Variables:
# mode: perl
# c-basic-offset:3
# End:
